@@ -27,6 +27,18 @@ let server_comments request t =
           parser
           every
 
+module Make_comp(C : Comparitor) : Comparitor_intf (* issue #7 *)
+       with type t := C.t
+        and module X = Z =
+  struct
+    type t = C.t
+    let ret = C.comp
+  end
+
+type 'a foo = 'a option =         (* Issue #98 *)
+  | None
+  | Some of 'a
+
 let qs1 = {| quoted string |}   (* (issue #24) *)
 let qs2 = {eof| other quoted string   |noteof}  |eof}
 
@@ -51,14 +63,19 @@ let x =
   | A -> 1                       (* Issue #73 *)
   end
 
+let command =                    (* Issue #130 *)
+  match x with
+    A ->
+     a
+  | B ->
+     b
+
 let x =
   let open M in
   let x = 5 in
   x + x
 ;;
 
-(* FIXME: MAJOR "function" sends SMIE into a loop (fine with "fun").
-   Use M-q to test. *)
 let () =
   let z = function t -> a in
   foo z
@@ -67,8 +84,6 @@ let () =
   foo(function t -> a)
 ;;
 
-(* FIXME: MAJOR: M-q on the "(" raises 'Scan error: "Unbalanced
-   parentheses"'.  It is fine if both () are on the same line. *)
 let () =
   begin
     (begin
@@ -155,8 +170,18 @@ type t = [                              (* Comment.  *)
   ]
 
 type t = a
- and typey = 4
- and x = b
+and typey = 4
+and x = b
+
+type t =
+  | A
+and u =
+  | B
+
+type _ gadt =
+  | A : int -> int gadt
+  | B : unit gadt
+  | C : float -> float gadt
 
 module M = struct
   type t =
@@ -476,6 +501,12 @@ let () =
      yyyyy
        (a b)
 
+let f = function
+    A ->
+     1
+  | B ->
+     2
+
 let d x = function
   (* FIXME: Should we leave it like this or align "|" with "match"?
      I chose with "match" because it looks otherwise odd and is more
@@ -507,6 +538,9 @@ let a f = function
 let x = foo ~f:(fun _ -> 0              (* Comment.  *)
                )
 
+let f = function x ->
+          y
+
 let f x =
   (let y = x in
    f x;
@@ -591,7 +625,7 @@ let () =
 let () =
   try f a
   with A () ->
-       ()
+        ()
      | B () ->
         ()
      |     C () ->
@@ -613,7 +647,7 @@ let () =
   | foo ->
      bar
      >>| function _ ->
-                  _
+       _
 
 let () =
   a
@@ -1145,13 +1179,13 @@ let x =
 let x =
   try a
   with Not_found ->
-       b
+        b
      | _ ->
         c
 let x =
   try a
   with Not_found ->
-       if a then b
+        if a then b
      | flag when String.is_prefix flag ~prefix:"-" ->
         a
      | _ ->
@@ -1274,9 +1308,9 @@ let subscribe_impl dir topic ~aborted =
 
 let x = List.map
           (function x ->
-                    blabla    (* FIXME: indentation afer "(function" *)
-                      blabla
-                      blabla)
+             blabla
+               blabla
+               blabla)
           l
 
 let command =
