@@ -31,73 +31,123 @@
   "Directory where to duplicate the files for flymake.")
 
 (defvar tuareg-jbuild-program
-  (expand-file-name "jbuilder-lint" tuareg-jbuild-temporary-file-directory)
+  (expand-file-name "jbuild-lint" tuareg-jbuild-temporary-file-directory)
   "Script to use to check the jbuild file.")
 
+(defgroup tuareg-jbuild nil
+  "Support for Jbuilder files."
+  :group 'languages)
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;                     Syntax highlighting
 
+(defface tuareg-jbuild-error-face
+  '((t (:foreground "yellow" :background "red" :bold t)))
+  "Face for errors (e.g. obsolete constructs).")
+
+(defvar tuareg-jbuild-error-face 'tuareg-jbuild-error-face
+  "Face for errors (e.g. obsolete constructs).")
+
 (defconst tuareg-jbuild-keywords-regex
-  (regexp-opt
-   '("jbuild_version" "library" "executable" "executables" "rule"
-     "ocamllex" "ocamlyacc" "menhir" "alias" "install" "copy_files")
-   'symbols)
+  (eval-when-compile
+    (concat (regexp-opt
+             '("jbuild_version" "library" "executable" "executables" "rule"
+               "ocamllex" "ocamlyacc" "menhir" "alias" "install"
+               "copy_files" "copy_files#" "include"
+               "documentation")
+             ) "\\(?:\\_>\\|[[:space:]]\\)"))
   "Keywords in jbuild files.")
 
 (defconst tuareg-jbuild-fields-regex
-  (regexp-opt
-   '("name" "public_name" "synopsis" "modules" "libraries" "wrapped"
-     "preprocess" "preprocessor_deps" "optional" "c_names" "cxx_names"
-     "install_c_headers" "modes" "no_dynlink" "kind"
-     "ppx_runtime_libraries" "virtual_deps" "js_of_ocaml" "flags"
-     "ocamlc_flags" "ocamlopt_flags" "library_flags" "c_flags"
-     "cxx_flags" "c_library_flags" "self_build_stubs_archive"
-     ;; + for "executable" and "executables":
-     "package" "link_flags" "modes" "names" "public_names"
-     ;; + for "rule":
-     "targets" "action" "deps" "fallback"
-     ;; + for "menhir":
-     "merge_into"
-     ;; + for "install"
-     "section" "files" "lib" "libexec" "bin" "sbin" "toplevel" "share"
-     "share_root" "etc" "doc" "stublibs" "man" "misc")
-   'symbols)
+  (eval-when-compile
+    (regexp-opt
+     '("name" "public_name" "synopsis" "modules" "libraries" "wrapped"
+       "inline_tests" "inline_tests.backend"
+       "preprocess" "preprocessor_deps" "optional" "c_names" "cxx_names"
+       "install_c_headers" "modes" "no_dynlink" "kind"
+       "ppx_runtime_libraries" "virtual_deps" "js_of_ocaml" "flags"
+       "ocamlc_flags" "ocamlopt_flags" "library_flags" "c_flags"
+       "cxx_flags" "c_library_flags" "self_build_stubs_archive"
+       "modules_without_implementation"
+       ;; + for "executable" and "executables":
+       "package" "link_flags" "modes" "names" "public_names"
+       ;; + for "rule":
+       "targets" "action" "deps" "mode"
+       ;; + for "menhir":
+       "merge_into"
+       ;; + for "install"
+       "section" "files" "lib" "libexec" "bin" "sbin" "toplevel" "share"
+       "share_root" "etc" "doc" "stublibs" "man" "misc"
+       ;; for "documentation":
+       "mld_files")
+     'symbols))
   "Field names allowed in jbuild files.")
 
-(defvar tuareg-jbuild-actions-regex
-  (regexp-opt
-   '("run" "chdir" "setenv"
-     "with-stdout-to" "with-stderr-to" "with-outputs-to"
-     "ignore-stdout" "ignore-stderr" "ignore-outputs"
-     "progn" "echo" "write-file" "cat" "copy" "copy#" "system" "bash")
-   t)
-  "Builtin actions in jbuilder")
+(defvar tuareg-jbuild-builtin-regex
+  (eval-when-compile
+    (concat (regexp-opt
+             '(;; Actions
+               "run" "chdir" "setenv"
+               "with-stdout-to" "with-stderr-to" "with-outputs-to"
+               "ignore-stdout" "ignore-stderr" "ignore-outputs"
+               "progn" "echo" "write-file" "cat" "copy" "copy#" "system"
+               "bash" "diff" "diff?"
+               ;; inline_tests and inline_tests.backend
+               ;; FIXME: "flags" is already a field and we do not have enough
+               ;; context to distinguishing both.
+               "backend" "generate_runner" "runner_libraries" "flags"
+               "extends"
+               ;; Dependency specification
+               "file" "alias" "alias_rec" "glob_files" "files_recursively_in"
+               "universe" "package")
+             t)
+            "\\(?:\\_>\\|[[:space:]]\\)"))
+  "Builtin sub-fields in jbuild")
 
 (defvar tuareg-jbuild-var-kind-regex
-  (regexp-opt
-   '("path" "path-no-dep" "exe" "bin" "lib" "libexec" "lib-available"
-     "version" "read" "read-lines" "read-strings")
-   'words)
+  (eval-when-compile
+    (regexp-opt
+     '("path" "path-no-dep" "exe" "bin" "lib" "libexec" "lib-available"
+       "version" "read" "read-lines" "read-strings")
+     'words))
   "Optional prefix to variable names.")
 
 (defvar tuareg-jbuild-var-regex
-      (concat "\\(\\(?:!\\|" tuareg-jbuild-var-kind-regex
-              ":\\)?\\)\\([a-zA-Z][a-zA-Z0-9_.]*\\|[<@^]\\)"
-              "\\(\\(?::[a-zA-Z][a-zA-Z0-9_.]*\\)?\\)"))
+      (concat "\\(!?\\)\\(\\(?:" tuareg-jbuild-var-kind-regex
+              ":\\)?\\)\\([a-zA-Z][a-zA-Z0-9_.-]*\\|[<@^]\\)"
+              "\\(\\(?::[a-zA-Z][a-zA-Z0-9_.-]*\\)?\\)"))
+
+(defmacro tuareg-jbuild--field-vals (field &rest vals)
+  `(list (concat "(" ,field "[[:space:]]+" ,(regexp-opt vals t))
+         1 font-lock-constant-face))
 
 (defvar tuareg-jbuild-font-lock-keywords
   `((,tuareg-jbuild-keywords-regex . font-lock-keyword-face)
-    (,tuareg-jbuild-fields-regex . font-lock-constant-face)
+    (,(concat "(" tuareg-jbuild-fields-regex) 1 font-lock-function-name-face)
     ("\\(true\\|false\\)" 1 font-lock-constant-face)
-    (,(concat "(" tuareg-jbuild-actions-regex) 1 font-lock-builtin-face)
+    ("(\\(select\\)[[:space:]]+[^[:space:]]+[[:space:]]+\\(from\\)\\>"
+     (1 font-lock-constant-face)
+     (2 font-lock-constant-face))
+    ,(eval-when-compile
+       (tuareg-jbuild--field-vals "kind" "normal" "ppx_rewriter" "ppx_deriver"))
+    ,(eval-when-compile
+       (tuareg-jbuild--field-vals "mode" "standard" "fallback" "promote"
+                                "promote-until-clean"))
+    (,(concat "(" tuareg-jbuild-builtin-regex) 1 font-lock-builtin-face)
+    ("(preprocess[[:space:]]+(\\(pps\\)" 1 font-lock-builtin-face)
+    (,(eval-when-compile
+        (concat "(" (regexp-opt '("fallback") t)))
+     1 tuareg-jbuild-error-face)
     (,(concat "${" tuareg-jbuild-var-regex "}")
-     (1 font-lock-builtin-face)
-     (3 font-lock-variable-name-face)
-     (4 font-lock-variable-name-face))
+     (1 tuareg-jbuild-error-face)
+     (2 font-lock-builtin-face)
+     (4 font-lock-variable-name-face)
+     (5 font-lock-variable-name-face))
     (,(concat "$(" tuareg-jbuild-var-regex ")")
-     (1 font-lock-builtin-face)
-     (3 font-lock-variable-name-face)
-     (4 font-lock-variable-name-face))
+     (1 tuareg-jbuild-error-face)
+     (2 font-lock-builtin-face)
+     (4 font-lock-variable-name-face)
+     (5 font-lock-variable-name-face))
     ("\\(:[a-zA-Z]+\\)\\b" 1 font-lock-builtin-face)))
 
 (defvar tuareg-jbuild-mode-syntax-table
@@ -113,6 +163,11 @@
     table)
   "Tuareg-jbuild syntax table.")
 
+;; (defun tuareg-jbuild-syntax-propertize (start end)
+;;     (funcall
+;;      (syntax-propertize-rules))
+;; )
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;                             SMIE
 
@@ -163,7 +218,7 @@
 (require 'flymake)
 
 (defun tuareg-jbuild-create-lint-script ()
-  "Create the lint script if it does not exist.  This is nedded as long as See https://github.com/janestreet/jbuilder/issues/241 is not fixed."
+  "Create the lint script if it does not exist.  This is nedded as long as See https://github.com/ocaml/dune/issues/241 is not fixed."
   (unless (file-exists-p tuareg-jbuild-program)
     (let ((dir (file-name-directory tuareg-jbuild-program))
           (pgm "#!/usr/bin/env ocaml
@@ -243,7 +298,7 @@ let () =
     files))
 
 (defun tuareg-jbuild--root (filename)
-  "Return the root and copy the necessary context files for jbuilder."
+  "Return the root and copy the necessary context files for jbuild."
   ;; FIXME: the root depends on jbuild-workspace.  If none is found,
   ;; assume the commands are issued from the dir where opam files are found.
   (let* ((dir (locate-dominating-file (file-name-directory filename)
@@ -294,7 +349,7 @@ let () =
   "Flymake entry for jbuild files.  See `flymake-allowed-file-name-masks'.")
 
 (defvar tuareg-jbuild--err-line-patterns
-  ;; Beware that the path from the root will be reported by jbuilder
+  ;; Beware that the path from the root will be reported by jbuild
   ;; but flymake requires it to match the file name.
   '(("File \"[^\"]*\\(jbuild\\)\", line \\([0-9]+\\), \
 characters \\([0-9]+\\)-\\([0-9]+\\): +\\([^\n]*\\)$"
@@ -316,8 +371,8 @@ characters \\([0-9]+\\)-\\([0-9]+\\): +\\([^\n]*\\)$"
   "(library" > \n
   "((name        " _ ")" > \n
   "(public_name " _ ")" > \n
-  "(synopsis  \"" _ "\")" > \n
-  "(libraries (" _ "))))" > ?\n)
+  "(libraries  (" _ "))" > \n
+  "(synopsis \"" _ "\")))" > ?\n)
 
 (define-skeleton tuareg-jbuild-insert-executable-form
   "Insert an executable stanza."
@@ -379,10 +434,18 @@ characters \\([0-9]+\\)-\\([0-9]+\\): +\\([^\n]*\\)$"
   nil
   "(copy_files " _ ")" > ?\n)
 
+(define-skeleton tuareg-jbuild-insert-documentation-form
+  "Insert a documentation stanza."
+  nil
+  "(documentation" > \n
+  "((package" _ ")" > \n
+  "(mld_files :standard)))" > ?\n)
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (defvar tuareg-jbuild-mode-map
   (let ((map (make-sparse-keymap)))
+    (define-key map "\C-c\C-c" 'compile)
     (define-key map "\C-c.v" 'tuareg-jbuild-insert-version-form)
     (define-key map "\C-c.l" 'tuareg-jbuild-insert-library-form)
     (define-key map "\C-c.e" 'tuareg-jbuild-insert-executable-form)
@@ -394,6 +457,7 @@ characters \\([0-9]+\\)-\\([0-9]+\\): +\\([^\n]*\\)$"
     (define-key map "\C-c.a" 'tuareg-jbuild-insert-alias-form)
     (define-key map "\C-c.i" 'tuareg-jbuild-insert-install-form)
     (define-key map "\C-c.c" 'tuareg-jbuild-insert-copyfiles-form)
+    (define-key map "\C-c.d" 'tuareg-jbuild-insert-documentation-form)
     map)
   "Keymap used in Tuareg-jbuild mode.")
 
@@ -425,6 +489,7 @@ characters \\([0-9]+\\)-\\([0-9]+\\): +\\([^\n]*\\)$"
   (setq-local comment-start ";")
   (setq-local comment-end "")
   (setq indent-tabs-mode nil)
+  ;(setq-local syntax-propertize-function #'tuareg-jbuild-syntax-propertize)
   (setq-local require-final-newline mode-require-final-newline)
   (push tuareg-jbuild--allowed-file-name-masks flymake-allowed-file-name-masks)
   (smie-setup tuareg-jbuild-smie-grammar #'tuareg-jbuild-smie-rules)
@@ -437,7 +502,7 @@ characters \\([0-9]+\\)-\\([0-9]+\\): +\\([^\n]*\\)$"
 
 ;;;###autoload
 (add-to-list 'auto-mode-alist
-             '("\\(?:\\`\\|/\\)jbuild\\'" . tuareg-jbuild-mode))
+             '("\\(?:\\`\\|/\\)jbuild\\(?:\\.inc\\)?\\'" . tuareg-jbuild-mode))
 
 
 (provide 'tuareg-jbuild-mode)
